"""
********************************************************************************
* Name: app_harvester
* Author: Nathan Swain and Scott Christensen
* Created On: August 19, 2013
* Copyright: (c) Brigham Young University 2013
* License: BSD 2-Clause
********************************************************************************
"""

import os
import inspect

from django.conf import settings
from sqlalchemy import create_engine

from tethys_apps.base import TethysAppBase


class SingletonAppHarvester(object):
    """
    Collects information for initiating apps
    """

    apps = []
    _instance = None

    def harvest_apps(self):
        """
        Searches the apps package for apps
        """
        # Notify user harvesting is taking place
        print('Harvesting Apps:')

        # List the apps packages in directory
        apps_dir = os.path.join(os.path.dirname(__file__), 'tethysapp')
        app_packages_list = os.listdir(apps_dir)

        # Harvest App Instances
        self._harvest_app_instances(app_packages_list)

        # Create Persistent stores
        self._provision_persistent_stores()
        #self._run_initialization_scripts()
        
    def __new__(self):
        """
        Make App Harvester a Singleton
        """
        if not self._instance:
            self._instance = super(SingletonAppHarvester, self).__new__(self)
            
        return self._instance

    @staticmethod
    def _validate_app(app):
        """
        Validate the app data that needs to be validated. Returns either the app if valid or None if not valid.
        """
        # Remove prepended slash if included
        if app.icon != '' and app.icon[0] == '/':
            app.icon = app.icon[1:]

        # Validate color
        if app.color != '' and app.color[0] != '#':
            # Add hash
            app.color = '#{0}'.format(app.color)

        # Must be 6 or 3 digit hex color (7 or 4 with hash symbol)
        if len(app.color) != 7 and len(app.color) != 4:
            app.color = ''

        return app

    def _harvest_app_instances(self, app_packages_list):
        """
        Search each app package for the app.py module. Find the AppBase class in the app.py
        module and instantiate it. Save the list of instantiated AppBase classes.
        """
        valid_app_instance_list = []
        
        for app_package in app_packages_list:
            # Collect data from each app package in the apps directory
            if app_package not in ['__init__.py', '__init__.pyc', '.gitignore', '.DS_Store']:
                # Create the path to the app module in the custom app package
                app_module_name = '.'.join(['tethys_apps.tethysapp', app_package, 'app'])

                # Import the app.py module from the custom app package programmatically
                # (e.g.: apps.apps.<custom_package>.app)
                app_module = __import__(app_module_name, fromlist=[''])
                
                for name, obj in inspect.getmembers(app_module):
                    # Retrieve the members of the app_module and iterate through
                    # them to find the the class that inherits from AppBase.
                    try:
                        # issubclass() will fail if obj is not a class
                        if (issubclass(obj, TethysAppBase)) and (obj is not TethysAppBase):
                            # Assign a handle to the class
                            _appClass = getattr(app_module, name)

                            # Instantiate app and validate
                            app_instance = _appClass()
                            validated_app_instance = self._validate_app(app_instance)

                            # compile valid apps
                            if validated_app_instance:
                                valid_app_instance_list.append(validated_app_instance)

                                # Notify user that the app has been loaded
                                print('{0}'.format(app_package))

                    except TypeError:
                        '''DO NOTHING'''
                    except:
                        raise

        # Save valid apps
        self.apps = valid_app_instance_list

    def _provision_persistent_stores(self):
        """
        Provision all persistent stores in the requested_stores property
        """
        # Notify user of database provisioning
        print('\nProvisioning Persistent Stores:')

        # Get database manager url from the config
        database_manager_url = settings.TETHYS_APPS_DATABASE_MANAGER_URL
        database_manager_name = database_manager_url.split('://')[1].split(':')[0]

        #--------------------------------------------------------------------------------------------------------------#
        # Get a list of existing databases
        #--------------------------------------------------------------------------------------------------------------#

        # Create connection engine
        engine = create_engine(database_manager_url)

        # Cannot create databases in a transaction: connect and commit to close transaction
        connection = engine.connect()

        # Check for Database
        existing_dbs_statement = '''
                                 SELECT d.datname as name
                                 FROM pg_catalog.pg_database d
                                 LEFT JOIN pg_catalog.pg_user u ON d.datdba = u.usesysid
                                 ORDER BY 1;
                                 '''

        existing_dbs = connection.execute(existing_dbs_statement)
        connection.close()

        # Compile list of db names
        existing_db_names = []

        for existing_db in existing_dbs:
            existing_db_names.append(existing_db.name)

        # Get apps and provision persistent stores if not already created
        for app in self.apps:
            # Create multiple persistent stores if necessary
            for persistent_store in app.persistent_stores():
                full_db_name = '_'.join((app.package, persistent_store.name))
                new_database = True

                #------------------------------------------------------------------------------------------------------#
                # 1. Create the database if it does not already exist
                #------------------------------------------------------------------------------------------------------#
                if full_db_name not in existing_db_names:
                    # Provide Update for User
                    print('Creating database "{0}" for app "{1}"...'.format(persistent_store.name, app.package))

                    # Cannot create databases in a transaction: connect and commit to close transaction
                    create_connection = engine.connect()

                    # Create db
                    create_db_statement = '''
                                          CREATE DATABASE {0}
                                          WITH OWNER {1}
                                          TEMPLATE template0
                                          ENCODING 'UTF8'
                                          '''.format(full_db_name, database_manager_name)

                    # Close transaction first and then execute
                    create_connection.execute('commit')
                    create_connection.execute(create_db_statement)
                    create_connection.close()

                else:
                    # Provide Update for User
                    print('Database "{0}" already exists for app "{1}", skipping...'.format(persistent_store.name,
                                                                                            app.package))

                    # Set var that is passed to initialization functions
                    new_database = False

                #------------------------------------------------------------------------------------------------------#
                # 2. Enable PostGIS extension
                #------------------------------------------------------------------------------------------------------#
                if persistent_store.postgis:
                    # Get URL for Tethys Superuser to enable extensions
                    super_url = settings.TETHYS_APPS_SUPERUSER
                    super_parts = super_url.split('/')
                    new_db_url = '{0}//{1}/{2}'.format(super_parts[0], super_parts[2], full_db_name)

                    # Connect to new database
                    new_db_engine = create_engine(new_db_url)
                    new_db_connection = new_db_engine.connect()

                    # Notify user
                    print('Enabling PostGIS on database "{0}" for app "{1}"...'.format(persistent_store.name,
                                                                                       app.package))
                    enable_postgis_statement = 'CREATE EXTENSION IF NOT EXISTS postgis'

                    # Execute postgis statement
                    new_db_connection.execute(enable_postgis_statement)
                    new_db_connection.close()

                #------------------------------------------------------------------------------------------------------#
                # 3. Run initialization function here
                #------------------------------------------------------------------------------------------------------#
                print('Initialize database "{0}" for app "{1}"'.format(persistent_store.name, app.package))

                # Split into module name and function name
                initializer_mod, initializer_function = persistent_store.initializer.split(':')

                # Pre-process initializer path
                initializer_path = '.'.join(('tethys_apps.tethysapp', app.package, initializer_mod))

                # Import module
                module = __import__(initializer_path, fromlist=[initializer_function])

                # Get the function
                initializer = getattr(module, initializer_function)
                initializer(new_database)

                #Spacer
                print('')



from tethys_apps.app_harvester import SingletonAppHarvester


def tethys_apps_context(request):
    """
    Add the current Tethys app metadata to the template context.
    """
    # Setup variables
    harvester = SingletonAppHarvester()
    context = {'tethys_app': None}
    apps_root = 'apps'

    # Get url and parts
    url = request.path
    url_parts = url.split('/')

    # Find the app key
    if apps_root in url_parts:
        # The app root_url is the path item following (+1) the apps_root item
        app_root_url_index = url_parts.index(apps_root) + 1
        app_root_url = url_parts[app_root_url_index]

        # Get list of app dictionaries from the harvester
        apps = harvester.apps

        # If a match can be made, return the app dictionary as part of the context
        for app in apps:
            if app.root_url == app_root_url:
                context['tethys_app'] = {'name': app.name,
                                         'index': app.index,
                                         'icon': app.icon,
                                         'color': app.color}
    return context


from django.shortcuts import render

from tethys_apps.app_harvester import SingletonAppHarvester


def library(request):
    """
    Handle the library view
    """
    # Retrieve the app harvester
    harvester = SingletonAppHarvester()

    # Define the context object
    context = {'apps': harvester.apps}

    return render(request, 'tethys_apps/app_library.html', context)



'''
********************************************************************************
* Name: New App Paste Template/Scaffold
* Author: Nathan Swain
* Copyright: (c) Brigham Young University 2013
* License: BSD 2-Clause
********************************************************************************
'''

import sys
import random
import re

from paste.script.templates import Template, var
from paste.util.template import paste_script_template_renderer
from paste.script.create_distro import Command

# Horrible hack to change the behaviour of Paste itself
# Since this module is only imported when commands are
# run, this will not affect any other paster commands.
Command._bad_chars_re = re.compile('[^a-zA-Z0-9_-]')


class TethysAppTemplate(Template):

    """
    Template to build a skeleton Tethys Apps package
    """

    _template_dir = 'tethysapp_template/'
    summary = 'Create a new Tethys app project using this scaffold.'
    template_renderer = staticmethod(paste_script_template_renderer)

    vars = [
        var('proper_name', 'e.g.: "My First App" for project name "my_first_app"'),
        var('version', 'e.g.: 0.0.1'),
        var('description', 'One-line description of the app'),
        var('author', 'Author name'),
        var('author_email', 'Author email'),
        var('url', 'URL of homepage'),
        var('license_name', 'License name'),
    ]

    # Default colors from flatuicolors.com
    default_colors = ('#2ecc71',    # Emerald
                      '#3498db',    # Peter River
                      '#34495e',    # Wet Asphalt
                      '#9b59b6',    # Amethyst
                      '#e67e22',    # Carrot
                      '#f1c40f',    # Sun Flower
                      '#e74c3c',    # Alizarin
                      '#1abc9c',    # Turquoise
    )

    def check_vars(self, vars, cmd):
        vars = Template.check_vars(self, vars, cmd)
        PREFIX = 'tethysapp-'
        PREFIX_NO_DASH = 'tethysapp'

        if not vars['project'].startswith(PREFIX):
            print('\nError: Expected the project name to start with "{0}". Please add the "{0}" '
                  'as a prefix and try again'.format(PREFIX))
            sys.exit(1)

        # Validate project name
        project_error_regex = re.compile(r'^[a-zA-Z0-9_]+$')
        project_warning_regex = re.compile(r'^[a-zA-Z0-9_-]+$')
        project = vars['project'][len(PREFIX):]

        # Only letters, numbers and underscores allowed in app names
        if not project_error_regex.match(project):

            # If the only offending character is a dash, replace dashes with underscores and notify user
            if project_warning_regex.match(project):
                before = project
                project = project.replace('-', '_')
                print('\nWarning: Dashes in project name "{0}" have been replaced ' \
                      'with underscores "{1}"'.format(before, project))

            # Otherwise, throw error
            else:
                print('\nError: Invalid characters in project name "{0}". Only letters, numbers, and underscores ' \
                      '(no dashes) allowed after the "tethysapp-" prefix.'.format(project))
                sys.exit(1)

        vars['project'] = project

        # Derive the project_url from the project name
        vars['project_url'] = project.replace('_', '-').lower()

        # Derive proper_name if not provided by user
        if not vars['proper_name']:
            vars['proper_name'] = project.replace('_', ' ').title()

        # Derive the proper_no_spaces variable (used for the name of the App class)
        vars['proper_no_spaces'] = ''.join(vars['proper_name'].split())

        # Add the color variable to vars
        vars['color'] = random.choice(self.default_colors)

        return vars



from .model import Base, engine, StreamGage, SessionMaker


def init_primary(first_time):
    """
    An example persistent store initializer function
    """

    # Create tables
    Base.metadata.create_all(engine)

    # First time add data
    if first_time:
        # Make a session
        session = SessionMaker()

        # Create StreamGage objects
        provo = StreamGage(name='Provo River Near Provo', lat=40.23833, lon=-111.6975)
        woodland = StreamGage(name='Lower River Near Woodland', lat=40.557778, lon=-111.181111)

        # Add to the session and commit
        session.add(provo)
        session.add(woodland)
        session.commit()


from django.conf.urls import patterns, url, include

from tethys_apps.utilities import generate_app_url_patterns

urlpatterns = patterns('',
    url(r'^$', 'tethys_apps.views.library', name='app_library'),
)

# Append the app urls urlpatterns
app_url_patterns = generate_app_url_patterns()

for namespace, urls in app_url_patterns.iteritems():
    root_pattern = r'^{0}/'.format(namespace.replace('_', '-'))
    urlpatterns.append(url(root_pattern, include(urls, namespace=namespace)))